import sys
import time
from asyncio import Future, AbstractEventLoop, Transport
from collections import deque, OrderedDict
from logging import Logger

import aioamqp
import asyncio
import contextlib
import contextvars
import datetime
import functools
import inspect
import logging
import os
import signal
import threading
import ujson
import uuid
from aioamqp import ChannelClosed, EmptyQueue, AmqpProtocol, AmqpClosedConnection
from aioamqp.channel import Channel
from aioamqp.envelope import Envelope
from aioamqp.properties import Properties
from pkg_resources import working_set
from typing import Optional, Callable, Any, Tuple, List, Dict, Awaitable, Union, Type, Generator
from typing_extensions import ContextManager, Coroutine  # 3.5 compat
from uuid import UUID

from garm_tools.modules.ctx import get_log, log_context, add_log, request_log_decorator
from garm_tools.modules.helpers import sys_log_print
from garm_tools.const import SUPERVISOR__CONTROL_STATE_FILE_DIR, SUPERVISOR__CONTROL_STATE_IS_ENABLED
from garm_tools.modules.logging import LoggerSteps

__all__ = ['pokemon_method', 'AbstractPokemon', 'AbstractRabbit', 'PokemonDeliveryError', 'check_session_reconnect',
           'pokemon_priority_method', 'claim_async_warn', 'logger', 'python_version']

UUID.__json__ = lambda self: f'"{self}"'
SEND_ALL_EXCHANGE = 'GlobalSendAll'

DELIVERY_MODE_REGULAR = 1
DELIVERY_MODE_PERSISTENT = 2

NO_RESULT = object()

CONTROL_STATE_UPDATE_TIMEDELTA = 10  # timedelta for update last access to state control file
CONTROL_STATE_MAX_LIFETIME = 30 * 24 * 60 * 60  # max lifetime of state control file


from ..tarantool import Tarantool

tarantool = Tarantool()

class PokemonDeliveryError(Exception):
    pass


def decode_body(body: Union[bytes, str]) -> str:
    if isinstance(body, bytes):
        # noinspection PyBroadException
        try:
            body = body.decode()
        except:  # noqa: E722
            body = str(body)
    return body


###############################################################################

def python_version(modules: bool = False) -> dict:
    """Check"""

    try:
        assert 1 == 0
        # noinspection PyUnreachableCode
        is_assert = False
    except AssertionError:
        is_assert = True

    ver = {
        'version': sys.version,
        'executable': sys.executable,
        'docs': python_version.__doc__ == 'Check',
        'assert': is_assert
    }
    if modules:
        # noinspection PyProtectedMember
        ver['modules'] = {d.project_name: d._version for d in working_set}

    return ver


###############################################################################

empty_coro = Future()
empty_coro.set_result(None)


# noinspection PyPep8Naming
class pokemon_priority_method:

    def __init__(self, priority: int = 0):
        self.priority = priority

    def __call__(self, func: Callable) -> Callable:
        claim_async_warn(func)

        @functools.wraps(func)
        async def decorated(decorated_self: 'AbstractPokemon', *args, **kwargs) -> Any:
            if not decorated_self.is_server:
                kwargs['priority'] = self.priority

            return await func(decorated_self, *args, **kwargs)

        return pokemon_method(decorated)


def real_execute_method(self: 'AbstractPokemon', coro: Awaitable) -> asyncio.Task:
    loop = asyncio.get_event_loop()

    queue = self.sync_queue
    task = loop.create_task(coro)
    queue.append(task)

    if not loop.is_running():
        while queue:
            current_task = queue.popleft()  # type: asyncio.Task
            loop.run_until_complete(current_task)

    return task


def get_call_ids() -> dict:
    log_ids = get_log()

    ids = {
        'page_id': log_ids.get('page_id', None),
        'req_id': log_ids.get('req_id', None)
    }
    if ids['req_id'] is None:
        ids['req_id'] = []

    return ids


def claim_async_warn(func: Callable) -> None:
    __annotations__ = getattr(func, '__annotations__', {})
    return_type = __annotations__.get('return')

    is_awaitable = inspect.iscoroutinefunction(func) or return_type == Coroutine
    if not is_awaitable:
        raise Exception(f'{func} is not coroutine and does not return coroutine')


def pokemon_method(func: Callable) -> Callable:  # noqa:CFQ004
    claim_async_warn(func)

    @functools.wraps(func)
    def method(self: 'AbstractPokemon', *args, get_coroutine: bool = False, **kwargs) -> Awaitable:  # noqa:CFQ004
        self._update_control_state_file()
        if self.is_server or self.sync_server:
            coro = func(self, *args, **kwargs)  # actual execution
            if not self.sync_server:
                return coro

            return real_execute_method(self, coro)

        elif self._batch_calls:
            self._batch_calls[-1].append((func.__name__, args, kwargs))
            loop = asyncio.get_event_loop()
            if self._can_create_loop_task and loop.is_running():
                return empty_coro
        else:
            return self.exec_self_action(
                func,
                *args,

                get_coroutine=get_coroutine,
                **kwargs
            )

    setattr(method, 'is_pokemon', True)

    return method


def generator_check_session_reconnect(func: Callable) -> Callable:
    from garm_tools.modules.logging.handlers import AbstractLogGarmHandler

    async def csr_decorated(self: 'AbstractRabbit', *args, **kwargs) -> bool:
        # noinspection PyBroadException
        try:
            step = 3
            while step:
                with AbstractLogGarmHandler.transaction():
                    try:
                        if not self.is_connected:
                            await self.connect()

                        async for res in func(self, *args, **kwargs):
                            yield res

                        return

                    except (ConnectionResetError, ChannelClosed):
                        AbstractLogGarmHandler.rollback()
                        self.logger.debug('Reconnect')
                        step -= 1
                        await asyncio.sleep(self._reconnect_time)

            raise PokemonDeliveryError
        except Exception as ex:
            self.logger.exception(self.__class__.__name__)

    return csr_decorated


def check_session_reconnect(func: Callable) -> Callable:
    from garm_tools.modules.logging.handlers import AbstractLogGarmHandler

    async def csr_decorated(self: 'AbstractRabbit', *args, real_result: bool = False, **kwargs) -> bool:
        # noinspection PyBroadException
        try:
            step = 3
            while step:
                with AbstractLogGarmHandler.transaction():
                    try:
                        if not self.is_connected:
                            await self.connect()

                        res = await func(self, *args, **kwargs)
                        return res if real_result else True
                    except (ConnectionResetError, ChannelClosed):
                        AbstractLogGarmHandler.rollback()
                        self.logger.debug('Reconnect')
                        step -= 1
                        await asyncio.sleep(self._reconnect_time)

            raise PokemonDeliveryError
        except Exception as ex:
            self.logger.exception(self.__class__.__name__)
            return False

    return csr_decorated


###############################################################################
class AbstractRabbit:
    _reconnect_time = 1

    __abstract__ = True

    is_server = False

    _local: contextvars.ContextVar[Optional[Channel]] = None
    _pid: int = None

    _loop: AbstractEventLoop = None

    _rabbit_host = 'localhost'
    _rabbit_port = None
    _rabbit_login = 'guest'
    _rabbit_passwd = 'guest'
    _login_method = 'PLAIN'
    _virtualhost = '/'
    _heartbeat_timeout = 60

    _default__queue_name: str = NotImplemented

    max_priority = 0

    logger: Logger = ...
    _log_type = 'rabbit'
    _control_state_file_path = None
    _last_access_control_state_file = 0

    _save_to_tarantool = False

    registry = {}

    def __init_subclass__(cls, scm_type=None, name=None, **kwargs):
        super().__init_subclass__(**kwargs)
        cls.registry[cls.__qualname__] = cls


    _info_col = 5
    _time_col = 6

    t = tarantool.add_scheme(
        'pokemon_status',
        fields=[
            ('pid', 'unsigned'),
            ('host', 'string'),
            ('name', 'string'),
            ('sv_group', 'string'),
            ('sv_name', 'string'),
            ('info', 'json'),
            ('start_time', 'unsigned')
        ],
        indexes=OrderedDict((
            ('hash', {'type': 'hash', 'parts': ['host', 'pid']}),  # удаление своей задачи
            ('name', {'type': 'tree', 'unique': False, 'parts': ['sv_group', 'name']}),  # все задачи одного покемона
            ('host_name', {'type': 'tree', 'unique': False, 'parts': ['host', 'sv_group', 'name']}),  # все задачи одного покемона на хосте
            ('host_sv_name', {'type': 'tree', 'unique': True, 'parts': ['host', 'sv_group', 'sv_name']}),  # предыдущий запуск на хосте
        ))
    )

    @functools.cached_property
    def pokemon_host_group_name(self) -> Tuple[str, str, str, str]:
        for cls in self.__class__.__mro__:
            if '_default__queue_name' in cls.__dict__ or '_queue_name' in cls.__dict__:
                break
        else:
            cls = self.__class__

        return (
            cls.__qualname__,
            os.uname()[1],
            os.environ.get('SUPERVISOR_GROUP_NAME', '-'),
            os.environ.get('SUPERVISOR_PROCESS_NAME', '-')
        )

    # noinspection TaskProblemsInspection
    def __init__(
            self,
            prefetch_count: int = 1,
            host: str = None,
            port: int = None,
            login: str = None,
            passwd: str = None,
            login_method: str = None,
            virtualhost: str = None,
            heartbeat_timeout: int = None,
            **kwargs
    ):
        if host:
            self._rabbit_host = host
        if port:
            self._rabbit_port = port
        if login:
            self._rabbit_login = login
        if passwd:
            self._rabbit_passwd = passwd
        if login_method:
            self._login_method = login_method
        if virtualhost:
            self._virtualhost = virtualhost
        if heartbeat_timeout:
            self._heartbeat_timeout = heartbeat_timeout

        self.logger = logging.getLogger(f'{self._log_type}.{self.__class__.__name__}')

        self._queue_name = kwargs.get('queue_name', self._default__queue_name)
        self._exchange = kwargs.get('exchange', '')

        self._prefetch_count = prefetch_count

        if SUPERVISOR__CONTROL_STATE_IS_ENABLED:
            self._control_state_file_dir = f"{SUPERVISOR__CONTROL_STATE_FILE_DIR}/{os.environ['SUPERVISOR_GROUP_NAME']}"
            self._control_state_file_name = f"{os.environ['SUPERVISOR_PROCESS_NAME']}.run"
            self._control_state_file_path = f'{self._control_state_file_dir}/{self._control_state_file_name}'

        self._ex = None
        self._message_result = None #type: Optional[dict]

    @property
    def channel(self) -> Optional[Channel]:
        # noinspection PyProtectedMember
        dest = self.channel_dest

        if dest._pid != (pid := os.getpid()):
            dest._local = threading.local()
            dest._pid = pid

        return getattr(dest._local, 'channel', None)

    @property
    def channel_dest(self) -> Type['AbstractRabbit']:
        # channel shared between all instances of some class
        return self.__class__

    @property
    def is_connected(self) -> bool:
        channel, transport, protocol = self.get_connection()

        if channel is None:
            return False

        return not transport.is_closing() and not protocol._connection_lost and channel.is_open

    def get_connection(self) -> Tuple[Optional[Channel], Optional[Transport], Optional[AmqpProtocol]]:
        dest = self.channel_dest
        # noinspection PyUnresolvedReferences, PyProtectedMember
        if (channel := self.channel) is None:
            return None, None, None
        return channel, getattr(dest._local, 'transport'), getattr(dest._local, 'protocol')

    def save_exception(self, ex: BaseException) -> None:
        if isinstance(ex, Exception):
            self._ex = ex

    async def basic_publish(self, payload: bytes, exchange_name: str, routing_key: str, properties: dict) -> None:
        self._ex = None
        await self.channel.basic_publish(
            payload=payload,
            exchange_name=exchange_name,
            routing_key=routing_key,
            properties=properties,
        )
        if self._ex is not None:
            raise self._ex

    async def connect(self) -> None:
        self.logger.debug('Connecting')
        dest = self.channel_dest

        while not self.is_connected:
            try:
                transport, protocol = await aioamqp.connect(
                    host=self._rabbit_host,
                    port=self._rabbit_port,
                    login=self._rabbit_login,
                    password=self._rabbit_passwd,
                    login_method=self._login_method,
                    on_error=self.save_exception,
                    virtualhost=self._virtualhost,
                    heartbeat=self._heartbeat_timeout
                )

                channel = await protocol.channel()

                dest._local.channel = channel
                dest._local.transport = transport
                dest._local.protocol = protocol

            except (OSError, aioamqp.exceptions.AmqpClosedConnection) as exc:
                sys_log_print(f'Rabbit connection error: {exc.__class__.__name__}: {exc}')
                await asyncio.sleep(self._reconnect_time)

            except Exception as exc:
                sys_log_print(f'Rabbit connect failed: {exc.__class__.__name__}: {exc}')
                raise exc

        if self.is_server:
            loop = asyncio.get_event_loop()  # create trap to auto-reconnect
            # noinspection PyAsyncCall
            loop.create_task(self._reconnect_cycle(self.channel.close_event))

        await self._declare_queue()

    def protected_loop(self, *args, **kwargs) -> AbstractEventLoop:
        """WorkController wrapper for create_task"""
        from .work_controller import MethodsWorkControl

        self._loop.create_task(MethodsWorkControl.protected_loop(
            *args, before_call=self._update_control_state_file, **kwargs
        ))

        return self._loop

    def prepare_server(self, user: Any = None) -> AbstractEventLoop:
        self.is_server = True

        self._loop = asyncio.get_event_loop()
        self._loop.create_task(self._receive())
        if self._control_state_file_path:
            self._run_control_state()

        _, host, sv_group, sv_name = self.pokemon_host_group_name
        if old_task := self.t.delete((host, sv_group, sv_name), index='host_sv_name'):
            message_result = old_task[0][self._info_col]
            if message_result and 'message_id' in message_result:
                self._message_result = message_result

        self._upsert_t('')

        return self._loop

    def run_server(self, user: Any = None) -> None:
        self.logger.info('Versions', extra={'info': ujson.dumps(python_version(modules=False))})
        self.prepare_server(user).run_forever()

    def _upsert_t(self, json: Any) -> tuple:
        name, host, sv_group, sv_name = self.pokemon_host_group_name
        cur_time = int(time.time())

        return self.t.upsert(
            (os.getpid(), host, name, sv_group, sv_name, json, cur_time),
            (('=', self._info_col, json), ('=', self._time_col, cur_time))
        )

    async def _reconnect_cycle(self, evt: asyncio.Event) -> None:
        await evt.wait()
        await self._receive()

    async def _declare_queue(self) -> dict:
        arguments = {}
        if self.max_priority:
            arguments['x-max-priority'] = self.max_priority

        if self._queue_name is NotImplemented:
            raise NotImplementedError(f'Need queue_name for {self.__class__.__name__}')

        # noinspection PyProtectedMember
        return await self.channel.queue_declare(
            queue_name=self._queue_name,
            durable=True,
            arguments=arguments,
        )

    async def _callback(self, channel: Channel, body: bytes, envelope: Envelope, properties: Properties) -> None:
        raise NotImplementedError

    async def _receive(self) -> None:
        await self.connect()
        await self.channel.basic_qos(prefetch_count=self._prefetch_count)
        await self.channel.basic_consume(
            self._callback,
            queue_name=self._queue_name,
        )
        self.logger.debug(f'Listening queue "{self._queue_name}"')

    def _run_control_state(self) -> None:
        """
        Запуск процесса контроля состояния покемона

        1. Проверка наличия каталога для хранения файлов состояния покемона
        2. Удаление старых файлов (время доступа > CONTROL_STATE_MAX_LIFETIME)
        3. Создание файла состояния текущего процесса покемона
        """
        os.makedirs(self._control_state_file_dir, exist_ok=True)

        for file_name in os.listdir(self._control_state_file_dir):
            file_path = f'{self._control_state_file_dir}/{file_name}'
            if time.time() - os.path.getatime(file_path) > CONTROL_STATE_MAX_LIFETIME:
                self.logger.info(f'remove control state file: {file_path}')
                os.remove(file_path)

        if not os.path.isfile(self._control_state_file_path):
            os.mknod(self._control_state_file_path)

        self._loop = asyncio.get_event_loop()
        self._loop.create_task(self._update_state())

    async def _update_state(self) -> None:
        """Цикл обновления времени доступа к файлу контроля состояния"""
        while True:
            self._update_control_state_file()
            await asyncio.sleep(1)

    def _can_update_control_state_file(self) -> bool:
        """Проверка возможности обновления времени доступа к файлу контроля состояния"""
        if self._control_state_file_path is not None:
            return time.time() - self._last_access_control_state_file > CONTROL_STATE_UPDATE_TIMEDELTA

    def _update_control_state_file(self) -> None:
        """Обновление времени доступа к файлу контроля состояния"""
        if self._can_update_control_state_file():
            os.utime(self._control_state_file_path)
            self._last_access_control_state_file = time.time()

    @generator_check_session_reconnect
    async def basic_get(self, ack: bool = None) -> Generator[dict, None, None]:
        try:
            channel = self.channel

            while True:
                obj = await channel.basic_get(self._queue_name)

                if ack is not None:
                    if ack:
                        await channel.basic_client_ack(delivery_tag=obj['delivery_tag'])
                    else:
                        await channel.basic_client_nack(delivery_tag=obj['delivery_tag'], requeue=False)

                yield obj

        except EmptyQueue:
            pass


class AbstractPokemon(AbstractRabbit):
    sync_server = False
    sync_queue = deque()  # queue we use instead of rabbitMQ's one for sync server
    need_klink = False

    error_delay = 1000  # ms
    error_delay_tries = 0  # int/'inf'
    error_nack = False  # True/False/'inf'

    error_delay_list = None  # type: List[Tuple[int, Union[int, str]]] # list of delays and their times

    _can_create_loop_task = True
    _delay_queues = set()

    _send_to_all_queue = False

    _responder = None
    _log_type = 'pokemon'
    defence = None  # type: Dict[str,Any]

    _need_stop = False
    _task_count = 0
    protected_exit_signals = {'SIGINT'}  # список сигналов для контроля завершения задач
    exit_signal_info = ...  # type: dict

    def __init__(  # noqa: CFQ002
            self,
            prefetch_count: int = 1,
            respond_to: 'AbstractPokemon' = None,
            host: str = None,
            port: int = None,
            login: str = None,
            send_to_roles: List[str] = None,
            passwd: str = None,
            delay: int = None,
            sync_server: bool = False,
            send_to_all: bool = False,
            **kwargs
    ):

        super().__init__(prefetch_count, host, port, login, passwd, **kwargs)
        self._respond_to = respond_to

        self.sync_server = sync_server
        self.delay = delay
        self._batch_calls = []
        if self._send_to_all_queue:
            self._send_to_roles = set(send_to_roles or [])
            self._send_to_roles.add(SEND_ALL_EXCHANGE)

        self.delay_queue_name = f'{self._queue_name}.delay.{delay}'

        if self.error_delay_list is None:
            self.error_delay_list = [(self.error_delay, self.error_delay_tries)]

        self._send_to_all = send_to_all
        self.current_try_number = 0

    @property
    def responder(self) -> 'Klink':
        if self._responder is None:
            self._responder = Klink(
                owner=self, prefetch_count=100,
                host=self._rabbit_host,
                port=self._rabbit_port,
                login=self._rabbit_login,
                passwd=self._rabbit_passwd,
                login_method=self._login_method,
                virtualhost=self._virtualhost,
                heartbeat=self._heartbeat_timeout
            )
        return self._responder

    @staticmethod
    def handle_exit_protected(protect_method: Callable) -> Callable:
        """
        Декоратор для защиты от преждевременного завершения
        """

        async def wrapper(self: 'AbstractPokemon', *args, **kwargs) -> Any:
            if self._need_stop:
                return

            try:
                self._task_count += 1
                return await protect_method(self, *args, **kwargs)
            finally:
                self._task_count -= 1
                self._stop_pokemon()

        return wrapper

    @staticmethod
    def _get_arg_kwarg(json: Dict) -> Tuple[List[Any], Dict[str, Any]]:
        return json.get('args', []), json.get('kwargs', {})

    async def execute(self, func: callable, json: Dict) -> Any:
        args, kwargs = self._get_arg_kwarg(json)
        return await func(*args, **kwargs)

    def exec_self_action(self, method: Callable, *args,
                         get_coroutine: bool = False,
                         respond_to: 'AbstractPokemon' = None,
                         send_to_all: bool = False,
                         **kwargs) -> Awaitable:
        try:
            loop = asyncio.get_event_loop()  # type: _UnixSelectorEventLoop
        except RuntimeError:
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)

        if respond_to is None:
            respond_to = self._respond_to

        task_uuid = respond_to and str(uuid.uuid4())

        coroutine = self.self_action(method.__name__, *args,
                                     send_to_all=send_to_all,
                                     respond_to=respond_to,  # too many kwargs
                                     task_uuid=task_uuid,  # too many kwargs
                                     **kwargs)

        if get_coroutine:
            return coroutine

        if AbstractPokemon._can_create_loop_task and loop.is_running():
            task = loop.create_task(coroutine)
        else:
            AbstractPokemon._can_create_loop_task = False
            loop.run_until_complete(coroutine)
            AbstractPokemon._can_create_loop_task = True
            task = None

        if respond_to:

            fut = asyncio.Future()

            if None:  # TODO some argument for that or get from method itself
                # getitem(method, 'timeout', None)
                timeout = 60.0
                fut = asyncio.wait_for(fut=fut, timeout=timeout)

            respond_to.responder.unresponded[task_uuid] = fut
            return fut

        return task

    @check_session_reconnect
    async def get_queue_info(self) -> dict:
        return await self.channel.queue_declare(queue_name=self._queue_name, durable=True)

    @request_log_decorator
    @check_session_reconnect
    async def self_action(self, method: str, *args,  # noqa:CFQ002
                          respond_to: 'Klink' = None,
                          task_uuid: str = None,
                          send_to_all: bool = False,
                          send_to_role: str = None,
                          priority: int = None,
                          **kwargs) -> None:
        self.logger.debug('Sending', extra={'method': method})

        routing_key = self._queue_name

        properties = {'message_id': str(uuid.uuid4())}

        if send_to_all or send_to_role or self._send_to_all:
            properties['delivery_mode'] = DELIVERY_MODE_REGULAR
            if respond_to:
                raise NotImplementedError('Cant send to all and get respond')
            if self.delay:
                raise NotImplementedError('Cant send to all and delay')

            exchange_name = send_to_role or SEND_ALL_EXCHANGE
        else:
            properties['delivery_mode'] = DELIVERY_MODE_PERSISTENT
            exchange_name = self._exchange
            if respond_to:
                responder = respond_to.responder
                if not responder.initialized:
                    # noinspection PyProtectedMember
                    await responder._receive()

                # noinspection PyProtectedMember
                properties['reply_to'] = responder._queue_name
                properties['correlation_id'] = task_uuid

            if self.delay:
                await self._declare_delay_queue(self.delay, self.delay_queue_name)

                routing_key = self.delay_queue_name
                properties['expiration'] = str(self.delay)

        if priority is not None:
            properties['priority'] = priority

        serializer = kwargs.pop('serializer', ujson.dumps)

        await self.basic_publish(
            payload=serializer({'method': method, 'log_ids': get_call_ids(), 'args': args, 'kwargs': kwargs}).encode(),
            exchange_name=exchange_name,
            routing_key=routing_key,
            properties=properties,
        )

        self.logger.debug('Sent', extra={'method': method})

    # noinspection PyUnresolvedReferences
    @contextlib.contextmanager
    def batch(self, ignore_errors: bool = False, repeat_failed_message: bool = False) -> ContextManager[asyncio.Future]:
        """
        Передача множества сообщений через RabbitMQ одним запросом.

        :param ignore_errors: Игнорирования ошибок. Если не включено, при возникновении ошибки - обработка всего запроса прервется.
        :param repeat_failed_message: При возникновении ошибки при обработке сообщения (любого из сообщений) - перезапускать его в работу.
        :return:
        """
        self._batch_calls.append([])

        fut = asyncio.Future()

        yield fut

        batch_calls = self._batch_calls.pop()
        if batch_calls:
            task = self.batch_call(*batch_calls, ignore_errors=ignore_errors, repeat_failed_message=repeat_failed_message)
            if task:
                task.add_done_callback(lambda t: fut.set_result(t.result()))
            else:
                fut.set_result(None)

    @pokemon_method
    async def batch_call(self, *args: Tuple[str, list, dict], ignore_errors: bool = False, repeat_failed_message: bool = False) -> None:
        for method_name, args2, kwargs2 in args:
            method = getattr(self, method_name)
            try:
                await method(*args2, **kwargs2)
            except Exception as exc:
                if ignore_errors or repeat_failed_message:
                    self.logger.exception('Ignore error in one message from batch and restart process'
                                          if repeat_failed_message else
                                          'Batch ignore')
                    if repeat_failed_message:
                        await getattr(self.__class__(), method_name)(*args2, **kwargs2)
                    continue
                raise exc

    def prepare_server(self, user: Any = None) -> AbstractEventLoop:
        """prepare server"""
        loop = super().prepare_server(user)

        for signame in self.protected_exit_signals:
            loop.add_signal_handler(
                getattr(signal, signame),
                functools.partial(self._handle_exit, signame, loop)
            )

        loop.create_task(self._stop_pokemon_task())

        return loop

    async def _declare_queue(self) -> None:

        await super()._declare_queue()

        for error_delay, _ in self.error_delay_list:
            await self._declare_delay_queue(error_delay, self._generate_error_queue_name(error_delay))

        await self._declare_delay_queue(self.delay, self.delay_queue_name)
        if self.is_server and self._send_to_all_queue:
            self._send_to_all_queue = await self._declare_send_all_queue()

    async def _declare_delay_queue(self, delay: int, delay_queue_name: str) -> Optional[str]:
        if not delay or delay_queue_name in self._delay_queues:
            return None

        self._delay_queues.add(delay_queue_name)

        await self.channel.exchange_declare(self._queue_name, 'fanout')
        await self.channel.queue_bind(self._queue_name, self._queue_name, '')

        # noinspection PyProtectedMember
        await self.channel.queue_declare(  # return to exchange self._queue_name after x-dead
            queue_name=delay_queue_name,
            durable=True,
            arguments={
                'x-dead-letter-exchange': self._queue_name
            }
        )

        return delay_queue_name

    async def _declare_send_all_queue(self) -> str:
        result = await self.channel.queue_declare(
            queue_name='',
            exclusive=True
        )
        queue_name = result['queue']

        for exchange_name in self._send_to_roles:
            await self.channel.exchange_declare(exchange_name=exchange_name, type_name='direct')
            await self.channel.queue_bind(
                exchange_name=exchange_name,
                queue_name=queue_name,
                routing_key=self._queue_name
            )

        await self.channel.basic_consume(
            callback=self._callback,
            queue_name=queue_name,
        )
        return queue_name

    async def _requeue(self, json: dict, envelope: Envelope, channel: Channel) -> Optional[bool]:
        finished = False
        tries = json.get('tries', 0)
        is_new = not envelope.is_redeliver and not tries

        curr_error_delay = self._get_error_delay(try_number=tries + 1)

        if self.error_nack and (is_new or self.error_nack == 'inf') and not AbstractPokemon.defence:
            await channel.basic_client_nack(delivery_tag=envelope.delivery_tag, requeue=True)
        elif curr_error_delay['curr_delay']:  # delays here
            if self._send_to_all_queue and envelope.exchange_name in self._send_to_roles:
                await channel.basic_client_nack(delivery_tag=envelope.delivery_tag, requeue=False)
                finished = None

            json['tries'] = tries + 1
            if AbstractPokemon.defence:
                json['defence'] = AbstractPokemon.defence
                AbstractPokemon.defence = None

            self.current_try_number = json['tries']

            publish_properties = {
                'delivery_mode': DELIVERY_MODE_PERSISTENT,
                'expiration': str(curr_error_delay['curr_delay']['delay'])
                # exception has no-priority
            }

            await self.basic_publish(
                payload=ujson.dumps(json).encode(),
                routing_key=self._generate_error_queue_name(curr_error_delay['curr_delay']['delay']),
                exchange_name='',
                properties=publish_properties
            )

            await channel.basic_client_nack(delivery_tag=envelope.delivery_tag, requeue=False)
        else:
            # all retry types failed: stop trying to process it
            finished = True  # should return exception to client
            await channel.basic_client_nack(delivery_tag=envelope.delivery_tag, requeue=False)

        return finished

    async def _do_reply_to(self, call_result: Any, properties: Properties) -> None:
        publish_properties = {
            'reply_to': properties.reply_to,
            'correlation_id': properties.correlation_id,
        }

        if call_result is NO_RESULT:
            call_result = None  # TODO maybe forward Error itself.

        if isinstance(call_result, BaseException):
            call_result = None  # TODO maybe forward Exception itself.

        call_result = ujson.dumps(call_result).encode()
        await self.basic_publish(
            payload=call_result,
            exchange_name='',
            routing_key=properties.reply_to,
            properties=publish_properties)

    async def _call_result(self, json: dict, properties: Properties) -> Any:
        if self._message_result and self._message_result['message_id'] == properties.message_id:
            exception = self._message_result.get('exception', None)
            if not exception and 'result' not in self._message_result:
                exception = True
                logger.exception('Daemon crushed on execution')

            return BaseException('Resend') if exception else self._message_result.get('result', None)

        if self._save_to_tarantool or self._message_result:
            json['message_id'] = properties.message_id
            self._upsert_t(json)
            self._message_result = None

        log_ids = json.get('log_ids') or {'req_id': []}
        add_log(**log_ids)

        AbstractPokemon.defence = json.get('defence') or {}

        with LoggerSteps('RUN', self.logger, extra={'method': json['method']}):
            func = getattr(self, json['method'], None)

            if getattr(func, 'is_pokemon', False):
                return await self.execute(func, json)
            else:
                self.logger.warning(f"No method '{json['method']}'")

        return NO_RESULT

    async def _send_result_to_rabbit(
            self, json: dict, call_result: Any, channel: Channel, envelope: Envelope, properties: Properties
    ) -> None:
        if json is None:
            await channel.basic_client_nack(delivery_tag=envelope.delivery_tag, requeue=False)
            finished = True
        elif isinstance(call_result, BaseException):
            finished = await self._requeue(json, envelope, channel)
            if finished is None:
                return

        else:  # everything is ok
            finished = True
            await channel.basic_client_ack(delivery_tag=envelope.delivery_tag)

        if properties.reply_to and finished:
            await self._do_reply_to(call_result, properties)

    async def _save_result_to_next_start(self, call_result: Any, properties: Properties):
        if properties.message_id is None:
            logger.error('no message_id in task')

        self._message_result = t_result = {
            'message_id': properties.message_id,
        }
        if isinstance(call_result, BaseException):
            t_result['exception'] = True
        else:
            t_result['result'] = call_result

        self._upsert_t(t_result)

    # this method was additional wrapped @handle_exit_protected decorator after class definition
    async def _callback(self, channel: Channel, body: bytes, envelope: Envelope, properties: Properties) -> None:
        json = None  # type: Optional[dict]

        self.logger.debug('Callback', extra={'info': decode_body(body)})

        with log_context():
            try:
                json = ujson.loads(body)
                call_result = await self._call_result(json, properties)
            except BaseException as e:
                call_result = e
                self.logger.exception(self.__class__.__name__)
            try:
                await self._send_result_to_rabbit(json, call_result, channel, envelope, properties)
            except AmqpClosedConnection:
                # Работает только в случае одновременно одного подписчика на очередь.
                await self._save_result_to_next_start(call_result, properties)
            else:
                if self._save_to_tarantool:
                    self._upsert_t('')

    async def _receive(self) -> None:
        if self.need_klink and self.responder and not self.responder.initialized:
            # noinspection PyProtectedMember
            await self.responder._receive()

        await super()._receive()

    def _generate_error_queue_name(self, error_delay: int) -> str:
        return f'{self._queue_name}.delay.{error_delay}'

    def _get_error_delay(self, try_number: int) -> Dict:
        """
        get current error delay, tries counter
        from self.error_delay_list with using try_number
        """
        curr_error_delay = None
        tries_counter = 0

        for error_delay in self.error_delay_list:
            if error_delay[1] != 'inf':
                tries_counter += error_delay[1]

            if error_delay[1] == 'inf' or try_number <= tries_counter:
                curr_error_delay = {'tries': error_delay[1], 'delay': error_delay[0]}
                break

        return {
            'curr_delay': curr_error_delay,
            'tries_counter': tries_counter,
        }

    # noinspection PyUnusedLocal
    def _handle_exit(self, signame: str, loop: AbstractEventLoop) -> None:
        """
        Установка флага о необходимости завершения
        работы покемона после возникновения сигнала из
        >>> self.protected_exit_signals
        """
        self._need_stop = True
        self.exit_signal_info = {
            'signame': signame,
            'received': datetime.datetime.now()
        }

    def _stop_pokemon(self) -> None:
        """
        Проверка необходимости и завершение работы покемона,
        если нет запущенных задач
        """
        if self._need_stop:
            if self._task_count > 0:
                self.logger.debug(
                    msg=f'Pokemon {self.__class__.__name__} is trying to stop, but for now it has running tasks'
                )
                return

            self.logger.info(
                msg=f"Pokemon {self.__class__.__name__} stopped with"
                    f" signal: {self.exit_signal_info['signame']}."
                    f" Signal received: {self.exit_signal_info['received']}"
            )
            exit()

    async def _stop_pokemon_task(self) -> None:
        """
        Задача вечной проверки на необходимость завершения работы покемона
        """
        while True:
            self._stop_pokemon()
            await asyncio.sleep(1)


# noinspection PyProtectedMember
AbstractPokemon._callback = AbstractPokemon.handle_exit_protected(AbstractPokemon._callback)


###############################################################################


class Klink(AbstractPokemon):

    def __init__(self, *args, owner: AbstractPokemon = None, **kwargs):
        super().__init__(*args, **kwargs)

        self.initialized = False

        self.unresponded = {}  # task_uuid to Future

        # noinspection PyProtectedMember
        self._queue_name = f'{owner._queue_name}_klink_{uuid.uuid4()}'

    @property
    def responder(self) -> 'Klink':
        return self

    @property
    def channel_dest(self) -> 'Klink':
        # unlike rest of classes, it stores channel for every instance, not class itself
        # otherwise 'SynchronizationError: Waiter already exists' raised
        # (or maybe there is just bad initialization flow)
        return self

    async def _declare_queue(self) -> dict:
        return await self.channel.queue_declare(queue_name=self._queue_name,
                                                exclusive=True, auto_delete=True)

    async def _callback(self, channel: Channel, body: bytes, envelope: Envelope, properties: Properties) -> None:
        fut = self.unresponded.pop(properties.correlation_id)
        result = ujson.loads(body)
        fut.set_result(result)

    async def _receive(self) -> None:
        self.initialized = True
        # set before we actually initialize that
        # to let other coroutines know there is someone working on that

        await self.connect()
        await self.channel.basic_qos(prefetch_count=self._prefetch_count)
        await self.channel.basic_consume(self._callback,
                                         queue_name=self._queue_name,
                                         no_ack=True)


logger = logging.getLogger('garm_tools')
