import asyncio
import datetime
import functools
import os
import signal
import time
from asyncio import AbstractEventLoop
from collections import OrderedDict
from typing import Optional, Callable, Coroutine

import psutil

from garm_tools.modules.ctx import infinite_req_log
from garm_tools.modules.pokemon import claim_async_warn, logger as gt_logger
from garm_tools.modules.pokemon.metrics import MonitMethodsWorkControl
from garm_tools.modules.tarantool import Tarantool, DatabaseError

__all__ = [
    'MethodsWorkControl',
    'MethodsWorkTimes'
]

WARN__MULTIPLIER = 5  # множитель проверки времени для отправки уведомлений о необычной длительности
ZOMBIE__MULTIPLIER = 100  # максимальный множитель проверки зомби
CHECK_AVG__COUNT = 10  # количество записей в таблице для получения среднего времени выполнения методов
MAX__WORK_TIME = 24 * 60 * 60  # максимальное время выполнения
MIN__WORK_TIME = 1  # минимальное время для проверки исполнения
ER_TUPLE_FOUND = 3  # SEE: garm_tools/modules/tarantool/error.py:_strerror ("Duplicate key exists in unique index '%s' in space '%s'")
HANDLE_SIGNALS = {
    'SIGABRT',

    # Используется в AbstractPokemon для контроля завершения запущенных задач
    # 'SIGINT',

    # Это невозможно отследить,
    # Не включать этот сигнал в список сигналов,
    # т.к. контролер перестает работать!!!
    # 'SIGKILL',
    # Оставляю этот сигнал как напоминание
    # Для сигнала 'kill' проходит сигнал 'SIGINT'

    'SIGHUP',
    'SIGTERM',
    'SIGBUS',
    'SIGILL',
    'SIGSEGV'
}

tarantool = Tarantool()
logger = gt_logger.getChild('methods_work_control')


class MethodsWorkTimes:
    """
    Контроль времени работы каждого конкретного метода.

    Механика:

    1. Ищется строка по method_name, берутся их work_time и усредняются -
    это и будет время для проверки, не завис ли процесс.
    Если в строке мало значений времени выполнения (< CHECK_AVG__COUNT),
    значит еще рано считать среднее и метод будет выполняться без контроля
    >>> get_work_time()

    2. После выполнения метода, нужно добавить его время выполнения в таблицу.
    Если в строке с method_name кол-во значений времени выполнения > CHECK_AVG__COUNT,
    значит новое время нужно добавить в конец, а первое - удалить
    >>> update_work_times()
    """

    __tablename__ = 'methods_work_times'

    t = tarantool.add_scheme(
        __tablename__,
        fields=[
            ('method_name', 'string'),
            ('work_times', 'json'),
            ('last_start_time', 'unsigned')
        ],
        indexes=OrderedDict((
            ('hash', {
                'type': 'hash',
                'parts': [
                    'method_name',
                ]
            }),
        ))
    )

    @classmethod
    def get_work_time(cls, method_name: str) -> int:
        """
        Получение среднего времени выполнения method_name.
        Если в таблице менее CHECK_AVG__COUNT записей с method_name -
        будет выдано среднее время * WARN__MULTIPLIER
        """
        work_time = MAX__WORK_TIME

        select_result = cls.t.select(method_name)

        if select_result:
            work_times = select_result.data[0][1]
            work_time = sum(work_times) // len(work_times) or MIN__WORK_TIME
            if len(work_times) < CHECK_AVG__COUNT:
                work_time = work_time * WARN__MULTIPLIER

        return int(work_time)

    @classmethod
    def update_work_times(cls, method_name: str, work_time: int, last_start_time: int) -> None:
        """
        Добавление времени работы методов в таблицу.
        Удаление старых результатов, если их количество > CHECK_AVG__COUNT
        """
        work_times = []
        select_result = cls.t.select(method_name)

        if select_result:
            work_times = select_result.data[0][1]

        work_times.append(work_time)
        work_times = work_times[1:CHECK_AVG__COUNT + 1] if len(work_times) > CHECK_AVG__COUNT else work_times

        cls.t.upsert(
            (method_name, work_times, last_start_time),
            [
                ('=', 'work_times', work_times),
                ('=', 'last_start_time', last_start_time)
            ]
        )


class StartWorkTime:
    """
    Получаем дату старта задачи для MethodsWorkControl
    """
    def __init__(self, hour: int = 0, minute: int = 0, second: int = 0, microsecond: int = 0, month_day: int = None):
        """
        :param hour:
        :param minute:
        :param second:
        :param microsecond:
        :param month_day: если задан, то отсчитываем дату в этом (или следующем) месяце
        """
        now = datetime.datetime.now()

        if month_day and now.day != month_day:
            calc_year = now.year
            calc_month = now.month

            if now.day > month_day:
                # поправка на декабрь
                if now.month == 12:
                    calc_year += 1
                    calc_month = 1
                else:
                    # поправка на количество дней в следующих месяцах
                    if now.month + 1 == 2:
                        is_leap = now.year % 4 == 0 and (now.year % 100 != 0 or now.year % 400 == 0)
                        month_day = min(29 if is_leap else 28, month_day)

                    elif now.month + 1 in {4, 6, 9, 11}:
                        month_day = min(30, month_day)

                    calc_month += 1

            self.start_time = now.replace(year=calc_year, month=calc_month, day=month_day, hour=hour,
                                          minute=minute, second=second, microsecond=microsecond).timestamp()
            return

        start_ms = hour * 60 * 60 * 1000 + minute * 60 * 1000 + second * 1000 + microsecond

        today_start = now.replace(hour=0, minute=0, second=0, microsecond=0)
        today_ms = (now - today_start).total_seconds() * 1000

        start_datetime = today_start if start_ms >= today_ms else (today_start + datetime.timedelta(hours=24))

        self.start_time = start_datetime.timestamp() + start_ms / 1000


class MethodsWorkControl:
    """
    Механизм защиты от запуска одного и того же метода в рамках разных демонов.
    В классе реализован декоратор work_decorator, со следующей механикой:
    1. производится поиск в таблице Tarantool записи по method.__name__
    2. если строка не найдена, то метод стартует
    3. если строка найдена, пытаемся запустить метод (если проходит проверка по ttl)
    4. после того, как метод отработал, удаляем строку из Tarantool
    """

    __tablename__ = 'methods_work_control'

    t = tarantool.add_scheme(
        __tablename__,
        fields=[
            ('method_name', 'string'),
            ('timestamp', 'unsigned'),
            ('ttl', 'unsigned'),
            ('send_warn', 'boolean'),
            ('hostname', 'string'),
            ('pid', 'unsigned')
        ],
        indexes=OrderedDict((
            ('hash', {
                'type': 'hash',
                'parts': [
                    'method_name',
                ]
            }),
            ('method_host', {
                'type': 'tree',
                'unique': True,
                'parts': [
                    'method_name',
                    'hostname',
                ]
            }),
            ('pid', {
                'type': 'hash',
                'parts': ['pid']
            })
        ))
    )

    @classmethod
    async def _work(cls, coro: Coroutine, ttl: int, hostname: str, start_time: int) -> bool:
        """
        Попытка старта работы передаваемой Coroutine,
        если разрешен ее запуск (проверка через insert)

        1. попытка вставки в таблицу строки с описанием информации
        об имени передаваемой Coroutine, времени старта и пр.
        2. если запуск разрешен, то пытаемся запустить coro и вернем True
        3. если запуск запрещен, возвращаем False
        """
        try:
            cls.t.insert((coro.__name__, start_time, ttl, False, hostname, os.getpid()))
        except DatabaseError as db_exc:
            if db_exc.args[0] == ER_TUPLE_FOUND:
                return False
            raise db_exc

        with MonitMethodsWorkControl.work_time.label(method=coro.__name__).timeit as work_time:
            try:
                MonitMethodsWorkControl.current.label(method=coro.__name__).set('1')
                await coro
            except Exception as ex:
                logger.exception('Error of running method "%s"' % coro.__name__, exc_info=ex)
            finally:
                MonitMethodsWorkControl.current.label(method=coro.__name__).set('0')
                cls.t.delete(coro.__name__)

        # сохраняем время выполнения метода в кеш
        MethodsWorkTimes.update_work_times(method_name=coro.__name__, work_time=work_time.val, last_start_time=start_time)
        return True

    @classmethod
    def _check_zombie_work(cls, method_name: str) -> Optional[int]:
        """
        Если в таблице еще найден процесс,
        который уже ZOMBIE__MULTIPLIER раз как должен был кончиться,
        то возвращаем время его старта
        """

        start_moment = False
        select_result = cls.t.select(method_name)

        if select_result:
            work_control = select_result.data[0]
            now_ts = int(time.time())

            # отправляем сообщение о том, что метод выполняется слишком долго
            if not work_control[3]:
                warn__method__end_time = work_control[1] + work_control[2] * WARN__MULTIPLIER
                if now_ts > warn__method__end_time:
                    MonitMethodsWorkControl.overttl.label(method=method_name).inc()

                    logger.warning(
                        'Method "%s" has been running for %s seconds (more than %s times longer than average)' %
                        (method_name, now_ts - work_control[1], WARN__MULTIPLIER)
                    )

                    cls.t.update(method_name, [('=', 'send_warn', True)], )

            guaranteed__method__end_time = work_control[1] + work_control[2] * ZOMBIE__MULTIPLIER
            if now_ts > guaranteed__method__end_time:
                start_moment = work_control[1]

        return start_moment

    @classmethod
    async def _zombie_work(cls, coro: Coroutine) -> Optional[bool]:
        """
        Пытаемся узнать, что метод уже должен был выполниться и умереть.
        Пытаемся его убить, если нужно (через удаление строки из кеша)
        Ожидаем возврат True только в том случае, если процесс не "завис"
        """
        check_zombie_work = cls._check_zombie_work(method_name=coro.__name__)
        if not check_zombie_work:
            coro.close()
            return True

        MonitMethodsWorkControl.killzombie.label(method=coro.__name__).inc()
        # todo kill zombie ?

        logger.error(
            'Detect zombie method call (%s, start time: %s, work_time: %s s)' %
            (coro.__name__, datetime.datetime.fromtimestamp(check_zombie_work), int(time.time() - check_zombie_work))
        )

        cls.t.delete(coro.__name__)

    @classmethod
    def work_controller(cls, method: Callable) -> Callable:
        """
        Декоратор для проверки состояния выполнения какого-то метода в данный момент времени
        """
        claim_async_warn(method)

        @functools.wraps(method)
        async def wrapper(*args, hostname: str, start_time: int, **kwargs) -> None:
            ttl = MethodsWorkTimes.get_work_time(method_name=method.__name__)

            coro = method(*args, **kwargs)  # type: Coroutine
            while not await cls._work(coro, ttl, hostname, start_time) and not await cls._zombie_work(coro):
                pass

        return wrapper

    @classmethod
    def _get_pid_process(cls, method_name: str, hostname: str = None) -> Optional[int]:
        if not hostname:
            hostname = os.uname()[1]

        work_control_cache = cls.t.select([method_name, hostname], index='method_host')
        if work_control_cache:
            return work_control_cache[0][5]

    @classmethod
    def _check_and_clean(cls, method_name: str, hostname: str = None) -> None:
        if not hostname:
            hostname = os.uname()[1]

        pid_process = cls._get_pid_process(method_name, hostname)
        if pid_process and not psutil.pid_exists(pid_process):
            cls.t.delete(method_name)

    @classmethod
    def __handle_exit(cls, signame: str, loop: AbstractEventLoop) -> None:
        print('WorkController (%s): was received signal %s' % (datetime.datetime.now(), signame), flush=True)
        curr_pid = os.getpid()
        work_cache = cls.t.select(curr_pid, index='pid')
        if work_cache:
            cls.t.delete(work_cache[0][0])
            print('WorkController (%s): killed pid %s process was removed from cache (method_name: %s)' %
                  (datetime.datetime.now(), curr_pid, work_cache[0][0]),
                  flush=True)
        loop.stop()

    @staticmethod
    def __get_last_start_time(method_name: str) -> int:
        work_times_cache = MethodsWorkTimes.t.select(method_name)
        last_start_time = work_times_cache[0][2] if work_times_cache.data else 0
        return last_start_time

    @classmethod
    async def protected_loop(cls, method: Callable, loop_time: float = 0, first_start: StartWorkTime = None,
                             before_call: Callable = None) -> None:
        """
        Метод для защиты от многократного запуска бесконечных задач

        :param method: Основная задача
        :param loop_time: Время от начала работы метода до следующего запуска
        :param first_start: Время первого запуска, в формате
        >>> StartWorkTime
        :param before_call: Метод для выполнения до основной задачи

        Пример использования:
        >>> def prepare_server(self, user: User = None) -> AbstractEventLoop:
        >>>     loop = super().prepare_server(user)
        >>>     # старый стиль
        >>>     loop.create_task(
        >>>         MethodsWorkControl.protected_loop(
        >>>             method=self._task,
        >>>             before_call=lambda x: None,
        >>>             loop_time=_task_period
        >>>         )
        >>>     )
        >>>     # с автоматической вставкой before_call=self._update_control_state_file:
        >>>     self.protected_loop(method=self._task, ...)
        >>>     return loop
        """

        hostname = os.uname()[1]
        cls._check_and_clean(method.__name__, hostname)

        loop = asyncio.get_event_loop()
        for signame in HANDLE_SIGNALS:
            loop.add_signal_handler(
                getattr(signal, signame),
                functools.partial(cls.__handle_exit, signame, loop)
            )

        wrapped_method = cls.work_controller(method)
        for _ in infinite_req_log():
            if before_call is not None:
                before_call()

            work_time = 0
            first_sleep_time = 0

            now_time = int(time.time())
            last_start_time = cls.__get_last_start_time(method.__name__)

            if not last_start_time and first_start is not None:
                first_sleep_time = first_start.start_time - now_time

            if now_time - last_start_time >= (first_start.start_time if first_sleep_time else loop_time):
                await wrapped_method(hostname=hostname, start_time=now_time)
                work_time = time.time() - now_time

            sleep_time = first_sleep_time or loop_time - work_time
            if sleep_time > 0:
                await asyncio.sleep(sleep_time)
